
.ENUM $C000
esc      DB
lzoffH:  DB
.ENDE        


; get byte from source data: hl = source position, bit 0 in c = nybble
.MACRO getbytefast
        ld      a,[hl+]
        bit     0,c
        jr      z,_done\@

        push    bc
        and     $0f                     ; current low nybble is actual hi
        ld      b,a
        ld      a,[hl]                  ; hi nybble in next pos is actual lo
        and     $f0
        or      b
        swap    a                       ; put nybbles in place
        pop     bc
_done\@:
.ENDM

; get nybble from source data: hl = source position, bit 0 in c = nybble
.MACRO getnybblefast
        inc     c                       ; always mirrors bit 0
        bit     0,c
        jr      nz,_hi\@

        ld      a,[hl+]                 ; get low nybble and advance hl
        and     $0f
        jr      _donef\@
_hi\@:
        ld      a,[hl]                  ; get high nybble
        swap    a
        and     $0f
_donef\@:
.ENDM


; hl = source (*.pag), de = destination

.SECTION "PacGirl - Unpagfast"

UnpagFast:
        ld      c,0                     ; bit 0 = nybble position

        call    getnybble               ; get the escape nybble used
        ld      [esc],a

unpagloop:
        ld      a,[esc]
        ld      b,a

        getnybblefast                   ; get one nybble from source
        cp      b
        jr      z,_escape

_noncompbyte:
        swap    a                       ; was high nybble of noncomp byte
        ld      b,a
        getnybblefast                   ; get low nybble of noncomp byte
        or      b

        ld      [de],a                  ; write out noncomp byte
        inc     de

        jr      unpagloop               ; move on

_lzshort:
        and     7                       ; low 3 bits are lz short length
        inc     a                       ; 1 + (0-7) bytes
        ld      b,a

        getnybblefast                   ; get lookback range (0-15)

        add     b                       ; calculate lookback offset
        cpl
        inc     a

        push    hl                      ; store source position

        add     e                       ; calculate lookback address
        ld      l,a                     ; for short lz
        ld      a,$ff
        adc     d
        ld      h,a

_lzloop:
        ld      a,[hl+]                 ; copy b bytes from earlier data
        ld      [de],a
        inc     de
        dec     b
        jr      nz,_lzloop

        pop     hl                      ; get source position from stack
        jr      unpagloop               ; move on

_compressed:
        bit     3,a                     ; info nybble 0nnn = lz short
        jr      z,_lzshort
        bit     2,a                     ; info nybble 10nn = lz long
        jr      z,_lzlong

        and     3                       ; 11nn = rle, nn = bit 5-4 of rle len
        swap    a
        ld      b,a
        call    getnybble               ; get low nybble of length
        or      b
        ret     z                       ; esc 1100 0000 = end mark

        add     2                       ; shortest run length is 1 + 2 bytes
        ld      b,a                     ; byte count in b

        getbytefast

_rleloop:
        ld      [de],a                  ; b bytes of byte value a
        inc     de
        dec     b
        jr      nz,_rleloop

        jr      unpagloop               ; move on

_escape:
        getnybblefast                   ; get info nybble after escape code
        or      a
        jr      nz,_compressed

        ld      a,[esc]                 ; old escape = actual high nybble
        ld      b,a

        call    getnybble               ; get new escape nybble
        ld      [esc],a

        ld      a,b                     ; high nybble, get low nybble
        jp      _noncompbyte

_lzlong:
        and     3                       ; low 2 bits = offset bits 10-9
        or      $fc
        ld      [lzoffH],a

        getnybblefast                   ; get lz length
        add     3
        ld      b,a                     ; 3 + (0-15) bytes

        getbytefast

        push    hl                      ; store source position

        add     e                       ; calculate lookback address
        ld      l,a                     ; for long lz
        ld      a,[lzoffH]
        adc     d
        ld      h,a

        jp      _lzloop


; get nybble from source data: hl = source position, bit 0 in c = nybble

getnybble:
        inc     c                       ; always mirrors bit 0
        bit     0,c
        jr      nz,_hinybble

        ld      a,[hl+]                 ; get low nybble and advance hl
        and     $0f
        ret
_hinybble:
        ld      a,[hl]                  ; get high nybble
        swap    a
        and     $0f
        ret

.ENDS
